// ==========================================================
// Lamprey_RK
// CBP 20-01-23 
// Orig code FBO 19-01-23
// ==========================================================

#include <Wire.h>
#include <Servo.h>

#define STATE_CALCULATE 1
#define STATE_TRANSMIT 2
#define STATE_READ 3

#define NEQ 1

#define DEBUG false

#define HEAD 1
#define SEGM 2
#define TAIL 3

int segType;

Servo testServo;
int servoSpeed;

int this_node_ID;
int target_node_ID;

int state;
bool endFound = false;

char message[20];
int i;

float u[NEQ];
float dudt[NEQ];
float yout[NEQ];

float deltaT;
float time;

float omega;
float phaseLag;
float k;
float tau;
float k1;
float tau1;
float input;

float omegaSegm;
float omegaHead;
float omegaTail;

float ad;
float delta;

int nrSegs;

void setup() {
  // put your setup code here, to run once:
  testServo.attach(13);
  testServo.write(90);

  pinMode(3, OUTPUT);
  Serial.begin(9600);

  deltaT = 0.01;
  time = 0;
  u[0] = 0.1;

  // Model parameters -----------------------------------------------

  omegaSegm = 2.0;
  delta = 60*PI/180.0;
  ad = 0.9;
  omegaHead = omegaSegm + ad*sin(delta);
  omegaTail = omegaSegm - ad*sin(delta);

  // Configuration --------------------------------------------------
/*
  segType = HEAD;
  this_node_ID = 0;
  target_node_ID = 1;
  state = STATE_CALCULATE;

  segType = SEGM;
  this_node_ID = 2;
  target_node_ID = this_node_ID + 1;
  state = STATE_READ;
  
  */

  segType = TAIL;
  this_node_ID = 3;
  target_node_ID = 0;
  state = STATE_READ;


  // ----------------------------------------------------------------

  Serial.print("segType = ");
  Serial.println(segType);
  Serial.print("This node = ");
  Serial.println(this_node_ID);
  Serial.print("Target node = ");
  Serial.println(target_node_ID);

  Wire.begin(this_node_ID);
  Wire.onReceive(receiveEvent);

}


void loop() {

  if (state == STATE_CALCULATE) {
    if(DEBUG) Serial.println("STATE_CALCULATE");
    solver_RK(NEQ, u, dudt, time, deltaT, yout, *rhs_segment);
    time += deltaT;
    for (i = 0; i < NEQ; i++) {
      u[i] = yout[i];
    }
    float op = sin(u[0]);
    servoSpeed = 180 * (op + 1.0) / 2;
    testServo.write(servoSpeed);
    dtostrf(u[0], 4, 1, message);
    int pwm = 100 + 100 * op;
    analogWrite(3, pwm);
    Serial.print("u[0] = ");
    Serial.print(u[0]);
    Serial.print("  op = ");
    Serial.println(message);
    state = STATE_TRANSMIT;
  }

  if (state == STATE_TRANSMIT) {
    if (DEBUG) Serial.print("STATE_TRANSMIT: Input = ");
    if (DEBUG) Serial.print(input);
    if (DEBUG) Serial.print(", Output (message) = ");
    if (DEBUG) Serial.println(message);

    Wire.beginTransmission(target_node_ID);
    Wire.print(message);
    Wire.print('E');
    byte status = Wire.endTransmission();
    if(status != 0x00) {
      Serial.println("Bus error");
    }

    state = STATE_READ;
  }

  if (state == STATE_READ) {
    //if (DEBUG)Serial.println("In STATE_READ");
    if (endFound) {
      state = STATE_CALCULATE;
      endFound = false;
    }
  }

}

void receiveEvent(int howMany) {
  if (DEBUG) Serial.println("receiveEvent from ");
  //Serial.println(target_node_ID);
  int i = 0;
  while (Wire.available() > 0) {
    char c = Wire.read(); // receive byte as a character
    if (DEBUG) Serial.println(c);
    if (c == 'E') {
      endFound = true;
      input = atof(message);
    } else {
       message[i] = c;
       endFound = false;
       i++;
    }
  }
}


void rhs_segment(float t, float u[], float dudt[]) {
  if(segType == HEAD){
    dudt[0] = omegaHead;
  }
  if(segType == SEGM){
    dudt[0] = omegaSegm + ad*sin( input - u[0] );
  }
  if(segType == TAIL){
    dudt[0] = omegaTail + ad*sin (input - u[0]);
  }
}

void solver_RK(int nVar, float y[], float* dydx, float t, float h, float* yout, void (*derivs) (float, float*, float*)) {
  float hh = h * 0.5;
  float h6 = h / 6.0;
  float xh = t + hh;
  float yt[NEQ];
  float dym[NEQ];
  float dyt[NEQ];

  derivs(t,y,dydx);

  for (int i = 0; i < nVar; i++) {
    yt[i] = y[i] + hh * dydx[i];
  }
  derivs(xh, yt, dyt);

  for (int i = 0; i < nVar; i++) {
    yt[i] = y[i] + hh * dyt[i];
  }
  derivs(xh, yt, dym);

  for (int i = 0; i < nVar; i++) {
    yt[i] = y[i] + h * dym[i];
    dym[i] += dyt[i];
  }
  derivs(t + h, yt, dyt);

  for (int i = 0; i < nVar; i++) {
    yout[i] = y[i] + h6 * (dydx[i] + dyt[i] + 2.0 * dym[i]);
  }
}
