// ==========================================================
// Myriapod_RK
// CBP 28-01-23
// New method of message passing allowing parameters to be 
// passed. These, when changed in the HEAD are passed down
// the whole line
// CBP 22-01-23 
// Orig code FBO 19-01-23
// ==========================================================

#include <Wire.h>

#define STATE_CALCULATE 1
#define STATE_TRANSMIT 2
#define STATE_READ 3

#define NEQ 6 //Largest required

#define DEBUG false

#define HEAD 1
#define SEGM 2
#define TAIL 3

#define LED_PIN 3

int segType;

int this_node_ID;
int target_node_ID;

int nrData;

int state;
bool endFound = false;

char message[20];
int i;

float u[NEQ];
float dudt[NEQ];
float yout[NEQ];

float deltaT;
float time;

// parameters for head, seg and tail
// head ----------------------------
float alpha;
float omega;
float mu;
// seg and tail---------------------
//float omega;
float phaseLag;
float k;
float tau;
float k1;
float tau1;
float input;

// Scaling of PWD for LED
float minVal;
float maxVal;

void setup() {

  pinMode(LED_PIN, OUTPUT);
  pinMode(8, OUTPUT);
  pinMode(7, INPUT);
  Serial.begin(9600);

  deltaT = 0.01;
  time = 0;

  // Model parameters -----------------------------------------------
  // TAIL and SEGM
  phaseLag = 60.0*(PI/180.0);
  // HEAD
  omega = 2*PI;
  alpha = 5;
  mu = 10;

  nrData = 5;

  // min and max vals for pwm scaling
  minVal = -1.0;
  maxVal =  1.0;

  // Configuration --------------------------------------------------
  segType = HEAD;
 // segType = SEGM;
 // segType = TAIL;

  if (segType == HEAD){
    this_node_ID = 0;
    target_node_ID = 1;
    state = STATE_CALCULATE;
  }

  if(segType == SEGM){
    this_node_ID = 3;
    target_node_ID = this_node_ID + 1;
    state = STATE_READ;
  }

  if(segType == TAIL){
    this_node_ID = 4;
    target_node_ID = 0;
   state = STATE_READ;
  }

  // Calc Initial Conditions ----------------------------------------

  if(segType == HEAD){
    u[0] = sqrt(2) / 2;
    u[1] = u[0];
  }
  if(segType == SEGM || segType == TAIL){
    setParams();
     for(int i = 0; i < 6; i++)
        u[i] = 0; 
  }

  // ----------------------------------------------------------------

  Serial.print("segType = ");
  Serial.println(segType);
  Serial.print("This node = ");
  Serial.println(this_node_ID);
  Serial.print("Target node = ");
  Serial.println(target_node_ID);

  Wire.begin(this_node_ID);
  Wire.onReceive(receiveEvent);

}

void setParams(){
     omega = omega + 1e-10;
	   tau = tan(phaseLag / 3) / omega;
	   k = sqrt(1 + tau * tau * omega * omega);
	   tau1 = tan(PI / 3) / omega;
	   k1 = sqrt(1 + tau1 * tau1 * omega * omega);
     //Serial.print("omega = ");Serial.println(omega);
     //Serial.print("phaselag = ");Serial.println(phaseLag);
}


void loop() {

  float op;
  if (state == STATE_CALCULATE) {
    if(DEBUG) Serial.println("STATE_CALCULATE");
    setParams();
    solver_RK(NEQ, u, dudt, time, deltaT, yout, *rhs_segment);
    time += deltaT;
    for (i = 0; i < NEQ; i++) {
      u[i] = yout[i];
    }
    if(segType == HEAD)
      op = u[0];
    else op = tanh(2.0*u[2]);
    int pwm = 255*(op - minVal)/(maxVal - minVal);
    if(pwm < 0) pwm = 0;
    if(pwm>255) pwm = 255;
    analogWrite(LED_PIN, pwm);
    if(DEBUG) Serial.print("u[0] = ");
    if(DEBUG) Serial.print(u[0]);
    if(DEBUG) Serial.print("  op = ");
    if(DEBUG) Serial.println(message);
    state = STATE_TRANSMIT;
  }

  if (state == STATE_TRANSMIT) {

    Wire.beginTransmission(target_node_ID);
    Wire.print('H');
    Wire.print(nrData);
    Wire.print(',');
    Wire.print(omega);
    Wire.print(',');
    Wire.print(phaseLag);
    Wire.print(',');
    Wire.print(minVal);
    Wire.print(',');
    Wire.print(maxVal);
    Wire.print(',');
    Wire.print(op);
    Wire.println(); // terminate message
    byte status = Wire.endTransmission();
    //Serial.println();
    if(status != 0x00) {
      Serial.println("Bus error");
    }
    delay(100);
    state = STATE_READ;
  }

  if (state == STATE_READ) {
    if (endFound) {
      state = STATE_CALCULATE;
      endFound = false;
    }
  }

}

void receiveEvent(int howMany) {
  if (DEBUG) Serial.println("receiveEvent from ");
  //Serial.println(target_node_ID);
  endFound = false;
  int i = 0;
  while (Wire.available() > 0) {
    char c = Wire.read(); // receive byte as a character
    if (DEBUG) Serial.println(c);
    if (c == 'H') {
      // first get nrData items
      int nrDataItems = Wire.parseInt();
      if (DEBUG) Serial.println(nrDataItems);      
      float float1 = Wire.parseFloat();
      omega = float1;
      float float2 = Wire.parseFloat();
      phaseLag = float2;
      float float3 = Wire.parseFloat();
      minVal = float3;
      float float4 = Wire.parseFloat();
      maxVal = float4;
      float float5 = Wire.parseFloat();
      input = float5;  

      if (DEBUG) { Serial.print("omega = "); Serial.println(float1); }
      if (DEBUG) { Serial.print("phaseLag = ");   Serial.println(float2);  }
      if (DEBUG) { Serial.print("val = "); Serial.println(float3); }
      endFound = true;
    }
  }
  //Serial.println();
}

void rhs_segment(float t, float u[], float dudt[]) {
  float r;
  if(segType == HEAD){
      r = sqrt(u[0] * u[0] + u[1] * u[1]);
      dudt[0] = alpha * (mu - r * r) * u[0] - omega * u[1];
      dudt[1] = omega * u[0] + alpha * (mu - r * r) * u[1];
  }

  if(segType == SEGM || segType == TAIL){
		// Left Leg
		dudt[0] = (-u[0] + k * input) / tau;
		dudt[1] = (-u[1] + k * u[0]) / tau;
		dudt[2] = (-u[2] + k * u[1]) / tau;

		// Right Leg
		dudt[3] = (-u[3] + k1 * u[2]) / tau1;
		dudt[4] = (-u[4] + k1 * u[3]) / tau1;
		dudt[5] = (-u[5] + k1 * u[4]) / tau1;
  }

}

void solver_RK(int nVar, float y[], float* dydx, float t, float h, float* yout, void (*derivs) (float, float*, float*)) {
  float hh = h * 0.5;
  float h6 = h / 6.0;
  float xh = t + hh;
  float yt[NEQ];
  float dym[NEQ];
  float dyt[NEQ];

  derivs(t,y,dydx);

  for (int i = 0; i < nVar; i++) {
    yt[i] = y[i] + hh * dydx[i];
  }
  derivs(xh, yt, dyt);

  for (int i = 0; i < nVar; i++) {
    yt[i] = y[i] + hh * dyt[i];
  }
  derivs(xh, yt, dym);

  for (int i = 0; i < nVar; i++) {
    yt[i] = y[i] + h * dym[i];
    dym[i] += dyt[i];
  }
  derivs(t + h, yt, dyt);

  for (int i = 0; i < nVar; i++) {
    yout[i] = y[i] + h6 * (dydx[i] + dyt[i] + 2.0 * dym[i]);
  }
}
