// =======================================================================
// Neural_Pulse_RK.ino
// CBP 17-12-22
// Pulse coupled segments with RK45 solver all in this file
// To do - place the solver in a library
// ========================================================================

#define HEAD 1
#define SEGM 2

unsigned long duration;
int thisNode;

unsigned long len;

float u0;

float u[6];
float dudt[6];
float yout[6];

float deltaT;
float time;

float alpha;
float omega;
float mu;

float tau;
float tau1;
float k;
float k1;
float phaseLag;
float input;

void setup() {

  Serial.begin(9600);
  
  pinMode(8, OUTPUT);
  pinMode(7, INPUT);

  pinMode(3, OUTPUT);

 // thisNode = HEAD;
 thisNode = SEGM;

  deltaT = 0.01;
  time = 0;

  omega = PI/2;
  alpha = 5;
  mu = 1;

  phaseLag = 180.0*(PI/180.0);

  if(thisNode == HEAD){
     init_head(alpha,omega,mu);
  } else if(thisNode == SEGM) {
    init_segm(omega,phaseLag);
  }

}

void transfering(unsigned long len) {
    
    digitalWrite(8, LOW);
    delayMicroseconds(100);
    digitalWrite(8, HIGH);
    delayMicroseconds(len);
    digitalWrite(8, LOW);
    delayMicroseconds(100);
}

void init_head(float _alpha, float _omega, float _mu){
  alpha = _alpha;
  omega = _omega;
  mu = _mu;
  u[0] = sqrt(2) / 2;
  u[1] = u[0];
}

void rhs_head(float t, float u[], float dudt[]) {
  float r;
  r = sqrt(u[0] * u[0] + u[1] * u[1]);
  dudt[0] = alpha * (mu - r * r) * u[0] - omega * u[1];
  dudt[1] = omega * u[0] + alpha * (mu - r * r) * u[1];

}

void init_segm(float _omega, float _phaseLag) {
  omega = _omega + 1e-10;
  phaseLag = _phaseLag;
	tau = tan(phaseLag / 3) / omega;
	k = sqrt(1 + tau * tau * omega * omega);
	tau1 = tan(PI / 3) / omega;
	k1 = sqrt(1 + tau1 * tau1 * omega * omega);

  for(int i = 0; i < 6; i++)
    u[i] = 0;
}

void segm_setInput(float _input) {
  input = _input;
}

void rhs_segm(float t, float u[], float dudt[]) {
		// Left Leg
		dudt[0] = (-u[0] + k * input) / tau;
		dudt[1] = (-u[1] + k * u[0]) / tau;
		dudt[2] = (-u[2] + k * u[1]) / tau;

		// Right Leg
		dudt[3] = (-u[3] + k1 * u[2]) / tau1;
		dudt[4] = (-u[4] + k1 * u[3]) / tau1;
		dudt[5] = (-u[5] + k1 * u[4]) / tau1;
    
}

void solver_RK(int nVar, float y[], float* dydx, float t, float h, float* yout, void (*derivs)(float, float*, float*)) {
  	float hh = h * 0.5;
		float h6 = h / 6.0;
		float xh = t + hh;
    float yt[6];
    float dym[6];
    float dyt[6];

    derivs(t,y,dydx);

		for (int i = 0; i < nVar; i++) {
			yt[i] = y[i] + hh * dydx[i];
		}
		derivs(xh, yt, dyt);

		for (int i = 0; i < nVar; i++) {
			yt[i] = y[i] + hh * dyt[i];
		}
		derivs(xh, yt, dym);

		for (int i = 0; i < nVar; i++) {
			yt[i] = y[i] + h * dym[i];
			dym[i] += dyt[i];
		}
		derivs(t + h, yt, dyt);
		for (int i = 0; i < nVar; i++) {
			yout[i] = y[i] + h6 * (dydx[i] + dyt[i] + 2.0 * dym[i]);
		}
}

int i;
long lastMillis = 0;
long loops = 0;

void loop() {

  if (thisNode == HEAD) {
 
    solver_RK(2, u, dudt, time, deltaT, yout, *rhs_head);
    time += deltaT;
    u[0] = yout[0];
    u[1] = yout[1];

    u0 = 100 + 100*yout[0];

    Serial.print(time);
    Serial.print(" ");
    Serial.println(u0);
    
    analogWrite(3, u0);
    transfering((unsigned long)u0);

  } else if (thisNode == SEGM) {

    duration = pulseIn(7, HIGH);
   
    float inputVal = ((float)duration - 100.0)/100.0;
    inputVal = tanh(2*inputVal);
    segm_setInput(inputVal);

    solver_RK(6, u, dudt, time, deltaT, yout, *rhs_segm);
    time += deltaT;
    u[0] = yout[0];
    u[1] = yout[1];
    u[2] = yout[2];

    float op = tanh(2*u[2]);
    float u2 = 100 + 100*op;
 

    Serial.print(input);
    Serial.print(" ");
    Serial.println(u[2]);
    analogWrite(3,u2);
    transfering((unsigned long)u2);

    i++;

  }

}